# mako/cache.py
# Copyright (C) 2006-2011 the Mako authors and contributors <see AUTHORS file>
#
# This module is part of Mako and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from bishop.mako import exceptions

cache = None

class BeakerMissing(object):
    def get_cache(self, name, **kwargs):
        raise exceptions.RuntimeException("the Beaker package is required to use cache functionality.")

class Cache(object):
    """Represents a data content cache made available to the module
    space of a :class:`.Template` object.
 
    :class:`.Cache` is a wrapper on top of a Beaker CacheManager object.
    This object in turn references any number of "containers", each of
    which defines its own backend (i.e. file, memory, memcached, etc.) 
    independently of the rest.
 
    """
 
    def __init__(self, id, starttime):
        self.id = id
        self.starttime = starttime
        self.def_regions = {}
 
    def put(self, key, value, **kwargs):
        """Place a value in the cache.
 
        :param key: the value's key.
        :param value: the value
        :param \**kwargs: cache configuration arguments.  The 
         backend is configured using these arguments upon first request.
         Subsequent requests that use the same series of configuration
         values will use that same backend.
 
        """
 
        defname = kwargs.pop('defname', None)
        expiretime = kwargs.pop('expiretime', None)
        createfunc = kwargs.pop('createfunc', None)
 
        self._get_cache(defname, **kwargs).put_value(key, starttime=self.starttime, expiretime=expiretime)
 
    def get(self, key, **kwargs):
        """Retrieve a value from the cache.
 
        :param key: the value's key.
        :param \**kwargs: cache configuration arguments.  The 
         backend is configured using these arguments upon first request.
         Subsequent requests that use the same series of configuration
         values will use that same backend.
 
        """
 
        defname = kwargs.pop('defname', None)
        expiretime = kwargs.pop('expiretime', None)
        createfunc = kwargs.pop('createfunc', None)
 
        return self._get_cache(defname, **kwargs).get_value(key, starttime=self.starttime, expiretime=expiretime, createfunc=createfunc)
 
    def invalidate(self, key, **kwargs):
        """Invalidate a value in the cache.
 
        :param key: the value's key.
        :param \**kwargs: cache configuration arguments.  The 
         backend is configured using these arguments upon first request.
         Subsequent requests that use the same series of configuration
         values will use that same backend.
 
        """
        defname = kwargs.pop('defname', None)
        expiretime = kwargs.pop('expiretime', None)
        createfunc = kwargs.pop('createfunc', None)
 
        self._get_cache(defname, **kwargs).remove_value(key, starttime=self.starttime, expiretime=expiretime)
 
    def invalidate_body(self):
        """Invalidate the cached content of the "body" method for this template.
 
        """
        self.invalidate('render_body', defname='render_body')
 
    def invalidate_def(self, name):
        """Invalidate the cached content of a particular <%def> within this template."""
 
        self.invalidate('render_%s' % name, defname='render_%s' % name)
 
    def invalidate_closure(self, name):
        """Invalidate a nested <%def> within this template.
 
        Caching of nested defs is a blunt tool as there is no
        management of scope - nested defs that use cache tags
        need to have names unique of all other nested defs in the 
        template, else their content will be overwritten by 
        each other.
 
        """
 
        self.invalidate(name, defname=name)
 
    def _get_cache(self, defname, type=None, **kw):
        global cache
        if not cache:
            try:
                from beaker import cache as beaker_cache
                cache = beaker_cache.CacheManager()
            except ImportError:
                # keep a fake cache around so subsequent 
                # calls don't attempt to re-import
                cache = BeakerMissing()

        if type == 'memcached':
            type = 'ext:memcached'
        if not type:
            (type, kw) = self.def_regions.get(defname, ('memory', {}))
        else:
            self.def_regions[defname] = (type, kw)
        return cache.get_cache(self.id, type=type, **kw)
 



import os.path
from bishop.mako.template import Template
from bishop.mako.lookup import TemplateLookup
import error, util, platform

class Role(object):
    def __init__(self, host, name):
        self.host = host
        self.name = name
        self._built = False
        self._order = None

        if not os.path.exists(self.path):
            raise error.BuildError('No role found with name %s' % name)

    @property
    def path(self):
        '''Return the on-disk path for this role'''
        return os.path.join(self.host.path, self.name)

    @property
    def mixins(self):
        '''Return the list of mixin names that are dependencies of this role'''
        try:
            f = open(os.path.join(self.path, 'mixins'), 'r')
            return util.uncomment_lines(f.read())
            f.close()
        except IOError:
            return []
    
    @property
    def packages(self):
        '''Return a list of package names that must be installed for this role to be installed'''
        try:
            f = open(os.path.join(self.path, 'packages'), 'r')
            return util.uncomment_lines(f.read())
            f.close()
        except IOError:
            return []
    
    def _relative_pathlist(self, dirname):
        '''Extract out the algorithm to return a list of files relative to a subdirectory of a role'''
        rel = []
        file_root = os.path.join(self.path, dirname)
        if not os.path.isdir(file_root):
            return rel

        for root, dirs, walk_files in os.walk(file_root):
            rel_root = root.replace(file_root, '')
            if rel_root.startswith('/'):
                rel_root = rel_root[1:]
            for file in walk_files:
                rel.append(os.path.join(rel_root, file))

            if len(walk_files) == 0 and len(dirs) == 0:
                '''Orphan directories will be added to the install just like files'''
                rel.append(rel_root)

        return rel

    @property
    def files(self):
        '''Return the relative (to the node-root) paths of all files for this role'''
        return self._relative_pathlist('files')

    @property
    def templates(self):
        '''Return the relative (to the node-root) paths of all templates for this role'''
        return self._relative_pathlist('templates')

    @property
    def partials(self):
        '''Return the relative (to the node-root) paths of all partials for this role'''
        return self._relative_pathlist('partials')

    @property
    def pre_install(self):
        preinst_path = os.path.join(self.path, 'preinst')
        if os.path.exists(preinst_path):
            return preinst_path
        return None
    
    @property
    def post_install(self):
        postinst_path = os.path.join(self.path, 'postinst')
        if os.path.exists(postinst_path):
            return postinst_path
        return None

    @property
    def pre_package(self):
        prepackage_path = os.path.join(self.path, 'prepackage')
        if os.path.exists(prepackage_path):
            return prepackage_path
        return None

    def build(self):
        '''Iterate through the role definition and signal to the host how to build the role data on the system'''
        if self._built:
            return

        for path in self.files:
            self.ready(path, link=os.path.join(self.path, 'files', path))

        lookup = TemplateLookup(directories=[os.path.join(self.path, 'templates'), os.path.join(self.path, 'template-includes')], strict_undefined=True)
        partial_lookup = TemplateLookup(directories=[os.path.join(self.path, 'partials'), os.path.join(self.path, 'template-includes')], strict_undefined=True)
        tpl_vars = util.get_tplvars()
        if 'cpu_count' not in tpl_vars: tpl_vars['cpu_count'] = util.cpu_count()
        if 'total_memory' not in tpl_vars: tpl_vars['total_memory'] = util.system_memory()
        if 'primary_ip' not in tpl_vars: tpl_vars['primary_ip'] = self.host.ip
        if 'hostname' not in tpl_vars: tpl_vars['hostname'] = platform.node()

        for path in self.templates:
            try:
                tpl = lookup.get_template(path)
                content = tpl.render(**tpl_vars)
            except Exception, e:
                raise error.TemplateError('%s in template %s/templates/%s' % (str(e), self.name, path))
            self.ready(path, content=content)

        for path in self.partials:
            try:
                tpl = partial_lookup.get_template(path)
                content = tpl.render(**tpl_vars)
                self.ready_partial(path, content)
            except Exception, e:
                raise error.TemplateError('%s in template %s/partials/%s' % (str(e), self.name, path))

        
        for package in self.packages:
            self.host.ready_package(package)
            if self.pre_package:
                self.host.ready_prepackage(self.pre_package, package)

        self._built = True

        for mixin in self.mixins:
            self.host.role(mixin).build()

    def ready(self, path, link=None, content=None):
        '''Fill the given path in the build directory with the given content or symlink'''
        build_path = os.path.join(self.host.build_path, path)
        
        if build_path in self.host._owned:
            conflict_role = self.host._owned[build_path]
            #Conflict resolution - if two roles are trying to get at the same file, the explicit role will win over the mixed-in role.  If both roles are mixed in or both are explicit, the conflict must be resolved by making one of them explicit and the other a mixin
            if conflict_role.name in self.host.roles and not self.name in self.host.roles:
                return
            elif self.name in self.host.roles and conflict_role.name not in self.host.roles:
                self.host._owned[build_path] = None
                os.remove(build_path)
            else:
                raise error.ConflictError('Both %s and %s are trying to install to "%s" - resolve by manually installing only one of the roles using "bishop install ROLENAME"' % (self.name, conflict_role.name, build_path))

        if self.pre_install:
            self.ready_meta(path, self.pre_install, 'preinstall')

        if self.post_install:
            self.ready_meta(path, self.post_install, 'postinstall')

        try:
            os.makedirs(os.path.dirname(build_path))
        except OSError:
            pass

        if link is not None:
            link_path = os.path.abspath(link)
            if not os.path.isdir(link_path) or not os.path.isdir(build_path):
                os.symlink(os.path.abspath(link), build_path)
        elif content is not None:
            f = open(build_path, 'w')
            f.write(content)
            f.close()
        else:
            raise error.BuildError('Trying to ready content at %s without a symlink or template' % path)

        self.host.ready(build_path, self)

    def ready_meta(self, path, meta_path, name):
        '''Attach a metadata file to a given path, tagged with a given name'''
        meta_root = os.path.join(self.host.build_path, '.bishop', os.path.dirname(path))
        try:
            os.makedirs(meta_root)
        except OSError:
            pass
        
        meta_build_path = os.path.join(meta_root, '%s-%s' % (os.path.basename(path), name))
        if os.path.exists(meta_build_path):
            os.remove(meta_build_path)

        os.symlink(os.path.abspath(meta_path), meta_build_path)

    def ready_partial(self, path, content):
        '''Take the rendered partial content and attach it to the file as metadata'''
        meta_root = os.path.join(self.host.build_path, '.bishop', os.path.dirname(path))
        try:
            os.makedirs(meta_root)
        except OSError:
            pass

        meta_build_path = os.path.join(meta_root, '%s-bpartial-%s' % (os.path.basename(path), self.name))
        f = open(meta_build_path, 'w')
        f.write(content)
        f.close()



import bishop
import os, os.path, shutil
from nose import tools

host = None

def setup_test():
    global host
    
    if os.path.isdir('test/test-install'):
        shutil.rmtree('test/test-install')
    
    for path in [os.path.join('test/test-repo', p) for p in os.listdir('test/test-repo') if p.startswith('.')]:
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)
    
    # Mercurial removes empty directories, but we explicitly need an empty directory for some of the tests
    if not os.path.isdir('test/test-repo/roledir/files/etc/empty-dir'):
        os.makedirs('test/test-repo/roledir/files/etc/empty-dir', 0755)
    if not os.path.isdir('test/test-repo/rolepostinst/files/etc/empty-dir'):
        os.makedirs('test/test-repo/rolepostinst/files/etc/empty-dir', 0755)
    
    host = bishop.host.Local(path='test/test-repo')

def teardown_test():
    global host
    host.unlock()
    host = None
    if os.path.isdir('test/test-install'):
        shutil.rmtree('test/test-install')
    
    for path in [os.path.join('test/test-repo', p) for p in os.listdir('test/test-repo') if p.startswith('.')]:
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)
    
    bishop.util.vars_path = '/etc/bishop-vars.json'
    bishop.util._tplvars = None


@tools.with_setup(setup_test, teardown_test)
def test_mirror():
    '''Every node will hold their own repository, but will only mirror the roles necessary to install on the node'''
    master_repo = bishop.repo.Repository(path=os.path.abspath('test/test-repo'))
    host = bishop.host.Local(path='test/test-repo/.copy')
    host.repo = master_repo

    host.mirror(['role1'])
    assert os.path.isdir('test/test-repo/.copy/role1')
    assert os.path.exists('test/test-repo/.copy/role1/mixins')
    assert os.path.isdir('test/test-repo/.copy/role1child')
    assert not os.path.isdir('test/test-repo/.copy/role2')
    assert not os.path.isdir('test/test-repo/.copy/.build')

@tools.with_setup(setup_test, teardown_test)
def test_build():
    '''Building a repository should create a .build directory in the repository root holding symlinks to all the files to install'''
    host.build(['role1', 'role2'])
    assert os.path.isdir('test/test-repo/.build')
    assert os.path.exists('test/test-repo/.build/rootfile')
    assert os.path.exists('test/test-repo/.build/etc/sample.conf')
    assert os.path.exists('test/test-repo/.build/etc/other_sample.conf')

@tools.with_setup(setup_test, teardown_test)
def test_build_templates():
    host.build(['role3'])
    assert os.path.exists('test/test-repo/.build/etc/templated.conf')
    assert open('test/test-repo/.build/etc/templated.conf').read().strip() == 'Hello World!'

@tools.with_setup(setup_test, teardown_test)
def test_build_packages():
    host.build(['role5'])
    assert open('test/test-repo/.build/.bishop/packages').read().strip() == 'lynx'

@tools.with_setup(setup_test, teardown_test)
def test_build_prepostinst():
    host.build(['role6'])
    assert os.path.exists('test/test-repo/.build/.bishop/etc/config.conf-postinstall')
    assert not os.path.exists('test/test-repo/.build/.bishop/etc/config.conf-preinstall')

@tools.with_setup(setup_test, teardown_test)
def test_build_node():
    host.roles = ['role2']
    host.build()
    assert os.path.exists('test/test-repo/.build/etc/sample.conf')

@tools.with_setup(setup_test, teardown_test)
def test_build_order():
    host.roles = ['order4', 'order3']
    host.build()
    assert host.ordered_roles == ['order1', 'order2', 'order3', 'order4']

    sorted_scripts = host.script_order({'test/test-repo/order1/postinst': [], 'test/test-repo/order2/postinst': []})
    assert('order1' in sorted_scripts[0][0] and 'order2' in sorted_scripts[1][0])



# mako/exceptions.py
# Copyright (C) 2006-2011 the Mako authors and contributors <see AUTHORS file>
#
# This module is part of Mako and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

"""exception classes"""

import traceback, sys, re
from bishop.mako import util

class MakoException(Exception):
    pass

class RuntimeException(MakoException):
    pass

def _format_filepos(lineno, pos, filename):
    if filename is None:
        return " at line: %d char: %d" % (lineno, pos)
    else:
        return " in file '%s' at line: %d char: %d" % (filename, lineno, pos)
 
 
class CompileException(MakoException):
    def __init__(self, message, source, lineno, pos, filename):
        MakoException.__init__(self, message + _format_filepos(lineno, pos, filename))
        self.lineno =lineno
        self.pos = pos
        self.filename = filename
        self.source = source
 
class SyntaxException(MakoException):
    def __init__(self, message, source, lineno, pos, filename):
        MakoException.__init__(self, message + _format_filepos(lineno, pos, filename))
        self.lineno =lineno
        self.pos = pos
        self.filename = filename
        self.source = source

class UnsupportedError(MakoException):
    """raised when a retired feature is used."""
 
class TemplateLookupException(MakoException):
    pass

class TopLevelLookupException(TemplateLookupException):
    pass
 
class RichTraceback(object):
    """Pulls the current exception from the sys traceback and extracts
    Mako-specific template information.
 
    See the usage examples in :ref:`handling_exceptions`.
 
    """
    def __init__(self, error=None, traceback=None):
        self.source, self.lineno = "", 0

        if error is None or traceback is None:
            t, value, tback = sys.exc_info()
 
        if error is None:
            error = value or t
 
        if traceback is None:
            traceback = tback
 
        self.error = error
        self.records = self._init(traceback)
 
        if isinstance(self.error, (CompileException, SyntaxException)):
            import mako.template
            self.source = self.error.source
            self.lineno = self.error.lineno
            self._has_source = True
 
        self._init_message()
 
    @property
    def errorname(self):
        return util.exception_name(self.error)
 
    def _init_message(self):
        """Find a unicode representation of self.error"""
        try:
            self.message = unicode(self.error)
        except UnicodeError:
            try:
                self.message = str(self.error)
            except UnicodeEncodeError:
                # Fallback to args as neither unicode nor
                # str(Exception(u'\xe6')) work in Python < 2.6
                self.message = self.error.args[0]
        if not isinstance(self.message, unicode):
            self.message = unicode(self.message, 'ascii', 'replace')

    def _get_reformatted_records(self, records):
        for rec in records:
            if rec[6] is not None:
                yield (rec[4], rec[5], rec[2], rec[6])
            else:
                yield tuple(rec[0:4])
 
    @property
    def traceback(self):
        """return a list of 4-tuple traceback records (i.e. normal python
        format) with template-corresponding lines remapped to the originating
        template.
 
        """
        return list(self._get_reformatted_records(self.records))
 
    @property
    def reverse_records(self):
        return reversed(self.records)
 
    @property
    def reverse_traceback(self):
        """return the same data as traceback, except in reverse order.
        """
 
        return list(self._get_reformatted_records(self.reverse_records))

    def _init(self, trcback):
        """format a traceback from sys.exc_info() into 7-item tuples,
        containing the regular four traceback tuple items, plus the original
        template filename, the line number adjusted relative to the template
        source, and code line from that line number of the template."""

        import mako.template
        mods = {}
        rawrecords = traceback.extract_tb(trcback)
        new_trcback = []
        for filename, lineno, function, line in rawrecords:
            if not line:
                line = ''
            try:
                (line_map, template_lines) = mods[filename]
            except KeyError:
                try:
                    info = mako.template._get_module_info(filename)
                    module_source = info.code
                    template_source = info.source
                    template_filename = info.template_filename or filename
                except KeyError:
                    # A normal .py file (not a Template)
                    if not util.py3k:
                        try:
                            fp = open(filename, 'rb')
                            encoding = util.parse_encoding(fp)
                            fp.close()
                        except IOError:
                            encoding = None
                        if encoding:
                            line = line.decode(encoding)
                        else:
                            line = line.decode('ascii', 'replace')
                    new_trcback.append((filename, lineno, function, line, 
                                            None, None, None, None))
                    continue

                template_ln = module_ln = 1
                line_map = {}
                for line in module_source.split("\n"):
                    match = re.match(r'\s*# SOURCE LINE (\d+)', line)
                    if match:
                        template_ln = int(match.group(1))
                    module_ln += 1
                    line_map[module_ln] = template_ln
                template_lines = [line for line in
                                    template_source.split("\n")]
                mods[filename] = (line_map, template_lines)

            template_ln = line_map[lineno]
            if template_ln <= len(template_lines):
                template_line = template_lines[template_ln - 1]
            else:
                template_line = None
            new_trcback.append((filename, lineno, function, 
                                line, template_filename, template_ln, 
                                template_line, template_source))
        if not self.source:
            for l in range(len(new_trcback)-1, 0, -1):
                if new_trcback[l][5]:
                    self.source = new_trcback[l][7]
                    self.lineno = new_trcback[l][5]
                    break
            else:
                if new_trcback:
                    try:
                        # A normal .py file (not a Template)
                        fp = open(new_trcback[-1][0], 'rb')
                        encoding = util.parse_encoding(fp)
                        fp.seek(0)
                        self.source = fp.read()
                        fp.close()
                        if encoding:
                            self.source = self.source.decode(encoding)
                    except IOError:
                        self.source = ''
                    self.lineno = new_trcback[-1][1]
        return new_trcback

 
def text_error_template(lookup=None):
    """Provides a template that renders a stack trace in a similar format to
    the Python interpreter, substituting source template filenames, line
    numbers and code for that of the originating source template, as
    applicable.
 
    """
    import mako.template
    return mako.template.Template(r"""
<%page args="error=None, traceback=None"/>
<%!
    from bishop.mako.exceptions import RichTraceback
%>\
<%
    tback = RichTraceback(error=error, traceback=traceback)
%>\
Traceback (most recent call last):
% for (filename, lineno, function, line) in tback.traceback:
  File "${filename}", line ${lineno}, in ${function or '?'}
    ${line | trim}
% endfor
${tback.errorname}: ${tback.message}
""")

def html_error_template():
    """Provides a template that renders a stack trace in an HTML format,
    providing an excerpt of code as well as substituting source template
    filenames, line numbers and code for that of the originating source
    template, as applicable.

    The template's default encoding_errors value is 'htmlentityreplace'. the
    template has two options. With the full option disabled, only a section of
    an HTML document is returned. with the css option disabled, the default
    stylesheet won't be included.
 
    """
    import mako.template
    return mako.template.Template(r"""
<%!
    from bishop.mako.exceptions import RichTraceback
%>
<%page args="full=True, css=True, error=None, traceback=None"/>
% if full:
<html>
<head>
    <title>Mako Runtime Error</title>
% endif
% if css:
    <style>
        body { font-family:verdana; margin:10px 30px 10px 30px;}
        .stacktrace { margin:5px 5px 5px 5px; }
        .highlight { padding:0px 10px 0px 10px; background-color:#9F9FDF; }
        .nonhighlight { padding:0px; background-color:#DFDFDF; }
        .sample { padding:10px; margin:10px 10px 10px 10px; font-family:monospace; }
        .sampleline { padding:0px 10px 0px 10px; }
        .sourceline { margin:5px 5px 10px 5px; font-family:monospace;}
        .location { font-size:80%; }
    </style>
% endif
% if full:
</head>
<body>
% endif

<h2>Error !</h2>
<%
    tback = RichTraceback(error=error, traceback=traceback)
    src = tback.source
    line = tback.lineno
    if src:
        lines = src.split('\n')
    else:
        lines = None
%>
<h3>${tback.errorname}: ${tback.message}</h3>

% if lines:
    <div class="sample">
    <div class="nonhighlight">
% for index in range(max(0, line-4),min(len(lines), line+5)):
    % if index + 1 == line:
<div class="highlight">${index + 1} ${lines[index] | h}</div>
    % else:
<div class="sampleline">${index + 1} ${lines[index] | h}</div>
    % endif
% endfor
    </div>
    </div>
% endif

<div class="stacktrace">
% for (filename, lineno, function, line) in tback.reverse_traceback:
    <div class="location">${filename}, line ${lineno}:</div>
    <div class="sourceline">${line | h}</div>
% endfor
</div>

% if full:
</body>
</html>
% endif
""", output_encoding=sys.getdefaultencoding(), encoding_errors='htmlentityreplace')



import os.path
import util, error

class AptGetPackage(object):
    path = None

    def install(self, name):
        try:
            return util.shell('export DEBIAN_FRONTEND=noninteractive && %s -qq -y install %s' % (self.path, name))
        except error.CalledProcessError, e:
            raise error.PackageError(str(e))

    def list(self):
        '''Return a list of all installed packages on the system'''
        packages = []
        for item in util.shell('dpkg --get-selections').strip().splitlines():
            packages.append(item.split()[0])

        return packages

    def extract(self, name, to_path):
        '''Download a given package along with its dependencies, then extract all files into the given folder'''
        util.shell('%s clean' % self.path)
        util.shell('%s -qq -y -d install %s' % (self.path, name))

        if not os.path.exists(to_path):
            os.makedirs(to_path, 0777)
        
        for path in [os.path.join('/var/cache/apt/archives/', path) for path in os.listdir('/var/cache/apt/archives') if path.endswith('.deb')]:
            util.shell('dpkg-deb -x %s %s' % (path, to_path))


class YumPackage(object):
    path = None

    def install(self, name):
        try:
            return util.shell('%s -q -y install %s' % (self.path, name))
        except error.CalledProcessError, e:
            raise error.PackageError(str(e))

    def list(self):
        '''Return a list of all installed packages on the system'''
        packages = []
        for item in util.shell('%s -q list installed' % self.path).strip().splitlines()[1:]:
            if item.startswith(' '): continue
            item = item.split()[0]
            if '.' in item: item = item.rsplit('.', 1)[0]
            packages.append(item)
        
        return packages

    def extract(self, name, path):
        raise NotImplementedError('auto-generating roles from packages is not supported on yum-based systems')

class UnknownPackage(object):
    '''If we can't find a package manager, allow Bishop to run but do not allow any package installation to take place'''
    def install(self, name):
        raise NotImplementedError('Package management is disabled, since yum and apt-get were not found')
    def list(self):
        raise NotImplementedError('Package management is disabled, since yum and apt-get were not found')
    def extract(self, name, path):
        raise NotImplementedError('Package management is disabled, since yum and apt-get were not found')


def search_path(bin):
    '''Given a binary name, search the environment PATH for a matching name'''
    path = os.getenv('PATH', '/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin')
    for location in path.split(os.pathsep):
        test_path = os.path.join(location, bin)
        if os.path.exists(test_path):
            return test_path

    raise ValueError('Could not find binary in path matching program name %s' % bin)


#Dynamically set Package class at time of import depending on the available binaries in the system
try:
    apt_path = search_path('apt-get')
    AptGetPackage.path = apt_path
    Package = AptGetPackage
except ValueError, e:
    try:
        yum_path = search_path('yum')
        YumPackage.path = yum_path
        Package = YumPackage
    except ValueError, e:
        Package = UnknownPackage





import SimpleXMLRPCServer

class Server(object):
    def __init__(self, host, port=4200):
        port = int(port)
        self.host = host
        self.server = SimpleXMLRPCServer.SimpleXMLRPCServer(('', port), allow_none=True)
        self.server.register_instance(self)

    def serve(self):
        self.server.serve_forever()

    def _dispatch(self, method, params):
        value = getattr(self.host, method)
        try:
            return value(*params)
        except TypeError:
            return value



import subprocess, logging, os.path, sys
import error

# Template vars will be pulled from /etc/bishop-vars.json as a JSON document, but of course a JSON library needs to be available
vars_enabled = False
vars_path = '/etc/bishop-vars.json'
try:
    import json
    vars_enabled = True
except:
    try:
        import simplejson as json
        vars_enabled = True
    except:
        pass

def shell(args):
    '''Wrapper around subprocess calls with a sensible default - return stdout as string.  On a non-zero error code, throw an exception with stderr'''
    logging.debug('shell: %s' % args)
    p = subprocess.Popen(args, shell=isinstance(args, basestring), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    retcode = p.wait()
    if retcode:
        raise error.CalledProcessError(retcode, args, stderr)

    return stdout

def uncomment_lines(body):
    '''Take a body string and return a list of lines with comments and empty lines stripped out'''
    def _uncomment_line(line):
        if '#' in line:
            return line[:line.index('#')].strip()
        return line.strip()

    return [_uncomment_line(line) for line in body.splitlines() if _uncomment_line(line) != '']

def chown_makedirs(path):
    '''Perform the same task as os.makedirs, but pull the uid and gid from the parent dir instead of the running process'''
    path = os.path.realpath(path)
    dirpath = '/'
    for part in path.strip('/').split('/'):
        dirpath = os.path.join(dirpath, part)
        if os.path.isdir(dirpath):
            continue
        if os.path.exists(dirpath):
            raise ValueError('%s exists but is not a directory' % dirpath)

        os.mkdir(dirpath)
        st = os.stat(os.path.dirname(dirpath))
        os.chown(dirpath, st.st_uid, st.st_gid)

_tplvars = None
def get_tplvars():
    global _tplvars

    if _tplvars is not None:
        return _tplvars
    if not os.path.exists(vars_path):
        _tplvars = {}
        return _tplvars
    if not vars_enabled:
        raise Exception('Bishop vars file requires Python 2.6+ or the simplejson library installed')
    _tplvars = json.load(open(vars_path, 'r'))
    return _tplvars

def validate_repopath(path):
    '''Test if a given path looks like a repo (all subfolders have files, templates, template-includes, or other bishop-isms)'''
    expect_dirs = ['files', 'templates', 'template-includes'] # at least one of these needs to appear for a role to be considered valid
    expect_files = ['packages', 'preinst', 'postinst', 'prepackage', 'mixins']

    for role in [role for role in os.listdir(path) if os.path.isdir(os.path.join(path, role)) and not role.startswith('.')]:
        role_path = os.path.join(path, role)
        good_role = False

        for d in expect_dirs:
            if os.path.isdir(os.path.join(role_path, d)):
                good_role = True

        for f in expect_files:
            if os.path.isfile(os.path.join(role_path, f)):
                good_role = True

        if not good_role:
            raise Exception('This isn\'t a bishop repo.  Are you trying to sync from some random directory?')
    
    return True

def get_repopath():
    '''Search the ancestors of the current working directory for a local bishop repo copy.  If we can't find one, the current working directory is the root of the copy and mark it.'''
    path = os.getcwd()
    while not os.path.exists(os.path.join(path, '.bishop.repo')) and path != '/':
        path = os.path.dirname(path)

    if path == '/':
        path = os.getcwd()

    mark_path = os.path.join(path, '.bishop.repo')
    if not os.path.exists(mark_path):
        validate_repopath(path)
        f = open(mark_path, 'w')
        f.write('1')
        f.close()

    return path

def cpu_count():
    '''Return the number of active CPU cores on this machine'''
    if 'bsd' in sys.platform or sys.platform == 'darwin':
        try:
            p = os.popen('sysctl -n hw.ncpu')
            return int(p.read())
        except ValueError:
            return 0
    else:
        try:
            return os.sysconf('SC_NPROCESSORS_ONLN')
        except (ValueError, OSError, AttributeError):
            return 0

def system_memory():
    '''Return the amount of RAM on the system in bytes'''
    return os.sysconf('SC_PHYS_PAGES') * os.sysconf('SC_PAGE_SIZE')



import bishop, os, shutil
from nose import tools

host = None

def setup_test():
    global host
    host = bishop.host.Local(path='test/test-repo')
    if os.path.isdir('test/test-install'):
        shutil.rmtree('test/test-install')

    # Mercurial removes empty directories, but we explicitly need an empty directory for some of the tests
    if not os.path.isdir('test/test-repo/roledir/files/etc/empty-dir'):
        os.makedirs('test/test-repo/roledir/files/etc/empty-dir', 0755)
    if not os.path.isdir('test/test-repo/rolepostinst/files/etc/empty-dir'):
        os.makedirs('test/test-repo/rolepostinst/files/etc/empty-dir', 0755)

def teardown_test():
    global host
    host.unlock()
    host = None
    if os.path.isdir('test/test-install'):
        shutil.rmtree('test/test-install')
    
    bishop.util.vars_path = '/etc/bishop-vars.json'
    bishop.util._tplvars = None


@tools.raises(bishop.error.PackageError)
@tools.with_setup(setup_test, teardown_test)
def test_badinstall():
    host.roles = ['role4']
    host.install(path='test/test-install')

@tools.with_setup(setup_test, teardown_test)
def test_goodinstall():
    host.roles = ['role1', 'role5']
    host.install(path='test/test-install')
    assert os.path.exists('test/test-install/etc/other_sample.conf')

@tools.with_setup(setup_test, teardown_test)
def test_installdir():
    '''An empty directory should still be created on install (eliminate need of superfluous .empty files)'''
    host.roles = ['roledir']
    host.install(path='test/test-install')
    assert os.path.exists('test/test-install/etc/empty-dir')

@tools.with_setup(setup_test, teardown_test)
def test_installdir_conflict():
    '''An empty directory that conflicts with a non-empty directory in a separate role should be maintained in the change list but ignored'''
    host.roles = ['roledir', 'roleconflictdir']
    host.install(path='test/test-install')
    assert os.path.exists('test/test-install/etc/empty-dir/not-empty')

@tools.with_setup(setup_test, teardown_test)
def test_orphan_postinstall():
    '''A postinstall script should be run on its own even if it has no files backing it - instead changes to the script itself will trigger rerun'''
    host.roles = ['rolepostinst']
    host.install(path='test/test-install')
    assert os.path.exists('/tmp/tmp-bishoptest-orphan-postinst')

@tools.with_setup(setup_test, teardown_test)
def test_bishopvars():
    '''A JSON object stored in /etc/bishop-vars.json should be implicitly used to create the context object for templates'''
    bishop.util.vars_path = 'test/test-repo/test-vars.json'
    bishop.util._tplvars = None
    
    host.roles = ['roletplvars']
    host.install(path='test/test-install')
    assert os.path.exists('test/test-install/test.txt')
    assert open('test/test-install/test.txt', 'r').read().strip() == 'Message: Hello World!'

@tools.with_setup(setup_test, teardown_test)
def test_partial():
    '''A role that claims a file and a partial for the file should include the content of both the file and the partial'''
    host.roles = ['partialrole']
    host.install(path='test/test-install')
    content = open('test/test-install/test.txt', 'r').read()
    assert content.strip() == 'files content\n\npartials content'

@tools.with_setup(setup_test, teardown_test)
def test_orphan_partial():
    '''A file that has a partial defined but no owning role should still contain the partials content'''
    host.roles = ['orphanpartial']
    host.install(path='test/test-install')
    content = open('test/test-install/test.txt', 'r').read()
    assert content.strip() == 'partials content'



import pickle, os.path, fcntl, time, shutil, logging, platform, xmlrpclib, sys, itertools
import role, package, util, error
try:
    from hashlib import sha1 as sha
except ImportError:
    from sha import new as sha

class Local(object):
    def __init__(self, path='/var/lib/bishop/'):
        self.path = path
        if not os.path.isdir(self.path):
            os.makedirs(self.path, 0777)
        self.repo = None
        self.last_install = 0
        self.install_locked = False
        self.roles = []
        self._install_roles = None
        self._roles = {}
        self._owned = {}
        self._locked = False
        self.load()

    def __del__(self):
        if self._locked: self.unlock()

    @property
    def build_path(self):
        return os.path.join(self.path, '.build')

    @property
    def name(self):
        return platform.node()

    @property
    def ip(self):
        '''Return the IP address that this host uses to connect to the repository'''
        if self.repo is None:
            return None

        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect((self.repo.hostname, 0))
        return s.getsockname()[0]

    @property
    def mixed_roles(self):
        '''Return the list of roles that this host implicitly installs via mixins from its explicit roles'''
        mixins = set()

        def recurse_mixins(role):
            for mixin in self.role(role).mixins:
                if not mixin in mixins:
                    mixins.add(mixin)
                    recurse_mixins(mixin)

        for role in self.roles:
            recurse_mixins(role)

        [mixins.discard(role) for role in self.roles]
        return list(mixins)

    @property
    def ordered_roles(self):
        '''Return the list of roles, including mixins, in installation order (deps first)'''
        self.mark_role_order()
        ordered = self.roles + self.mixed_roles
        ordered.sort(key=lambda x: -self.role(x)._order)
        return ordered
    
    def load(self):
        '''Load the saved host data from the serialized snapshot file'''
        save_path = os.path.join(self.path, '.save')
        if not os.path.exists(save_path):
            return

        saved = pickle.load(open(save_path, 'r'))
        for k, v in saved.iteritems():
            setattr(self, k, v)

    def save(self):
        '''Save this host's current options to a serialized snapshot file'''
        unlock = self.lock()
        
        to_save = ['repo', 'roles', 'last_install', 'install_locked']
        saved = {}
        for attr in to_save:
            saved[attr] = getattr(self, attr)

        save_path = os.path.join(self.path, '.save')
        f = open(save_path, 'w')
        pickle.dump(saved, f)
        f.close()

        if self.repo is not None:
            self.repo.save(self)

        if unlock: self.unlock()

    def mirror(self, roles=None, done=None, repo_roles=None):
        '''Find the roles given in this host's repository and sync changes down to the host cache'''
        if repo_roles is None: repo_roles = self.repo.all_roles
        
        if roles is None:
            self.roles = [role for role in self.roles if role in repo_roles]
            roles = self.roles
        elif len(roles) > 0:
            roles = [role for role in roles if role in repo_roles]

        if len(roles) == 0:
            return
        
        # Accumulate the finished roles so we don't double-sync or fail on circular mixins
        if done is None:
            done = {}

        mirror_roles = [role for role in roles if role not in done]
        if len(mirror_roles) == 0:
            return

        self.repo.copy_down('', self.path, ['/' + role + '/' for role in mirror_roles] + ['/' + role + '/**' for role in mirror_roles])
        
        mixin_roles = []
        for role in mirror_roles:
            done[role] = True
            mixin_roles.extend([mix for mix in self.role(role).mixins if mix not in done and mix not in mixin_roles])

        self.mirror(mixin_roles, done, repo_roles)
    
    def role(self, name):
        '''Given a name, make sure we have a Role instance, cached by name per host'''
        if isinstance(name, role.Role):
            return name

        if name not in self._roles:
            self._roles[name] = role.Role(self, name)

        return self._roles[name]
    
    def mark_role_order(self, mark_roles=None):
        '''Go through all roles and mark the order field based on the role's installation depth'''
        def recurse_order(roles, order=1):
            if len(roles) == 0:
                return

            for name in reversed(roles):
                self.role(name)._order = order
                order += 1
            recurse_order(list(itertools.chain(*[self.role(name).mixins for name in roles])), order)

        if mark_roles is None:
            mark_roles = self.roles
        recurse_order(mark_roles)

    def sync_down(self):
        '''Copy from the repository to the current working directory'''
        disabled = self.repo.sync_disabled
        if disabled:
            print disabled
            return
        path = util.get_repopath()
        repo_roles = self.repo.all_roles
        self.repo.copy_down('', path + '/', ['/' + role + '/' for role in repo_roles] + ['/' + role + '/**' for role in repo_roles])
        self.snapshot_wc(path)

    def sync_up(self):
        '''Copy from the current working directory up to the repository'''
        disabled = self.repo.sync_disabled
        if disabled:
            print disabled
            return
        path = util.get_repopath()
        roles = [role for role in os.listdir(path) if not role.startswith('.') and os.path.isdir(os.path.join(path, role))]
        
        snap_path = os.path.join(path, '.bishop.files')
        try:
            f = open(snap_path, 'r')
            prev_files = pickle.load(f)
            f.close()
        except:
            prev_files = {}

        for role in roles:
            changed, deleted = self.sync_changes(os.path.join(path, role), prev_files)
            
            if len(deleted) > 0:
                #Deletes cannot be rsynced, so just sync the whole changed role
                self.repo.copy_up(os.path.join(path, role) + '/', role + '/')
            else:
                for change_file in changed:
                    self.repo.copy_up(change_file, change_file[len(path) + 1:])

        # if we succeed in syncing up to the repo, remove any locks on remote install since local and remote are in sync
        if self.install_locked:
            self.install_locked = False
            self.save()

        self.snapshot_wc(path)

    def snapshot_wc(self, path):
        '''Take a snapshot of the files and their modification times so we can be more targetted in what we sync back up to the repo - some multi-user protection'''
        roles = [role for role in os.listdir(path) if not role.startswith('.') and os.path.isdir(os.path.join(path, role))]
        
        snapshot = {}
        for role in roles:
            for root, dirs, files in os.walk(os.path.join(path, role)):
                for fpath in files:
                    full_path = os.path.join(root, fpath)
                    snapshot[full_path] = os.stat(full_path).st_mtime

                for dpath in dirs:
                    full_path = os.path.join(root, dpath)
                    snapshot[full_path] = os.stat(full_path).st_mtime

                if len(files) == 0 and len(dirs) == 0 and root not in snapshot:
                    #Orphan directories should be snapshotted
                    snapshot[root] = os.stat(root).st_mtime
        
        snap_path = os.path.join(path, '.bishop.files')
        f = open(snap_path, 'w')
        pickle.dump(snapshot, f)
        f.close()
    
    def sync_changes(self, path, prev_files):
        '''Compare the current tree with the previous snapshot to get a list of changed files within a given role'''
        
        new_role = True
        found_files = []
        changed = []
        for root, dirs, files in os.walk(path):
            rel_root = root[len(path) - 1:]
            for fpath in files:
                full_path = os.path.join(root, fpath)
                found_files.append(full_path)
                if full_path not in prev_files:
                    changed.append(full_path)
                elif os.stat(full_path).st_mtime > prev_files[full_path]:
                    changed.append(full_path)
                    new_role = False
                else:
                    new_role = False

            for dpath in dirs:
                #All sub directories should be tracked
                full_path = os.path.join(root, dpath)
                found_files.append(full_path)
                if full_path not in prev_files:
                    changed.append(full_path + '/')
                else:
                    new_role = False

            if len(files) == 0 and len(dirs) == 0:
                #New orphan directories should also be tracked, if they aren't already tracked
                if root not in found_files:
                    found_files.append(root)
                    if '/' in rel_root and root not in prev_files:
                        changed.append(root + '/')
                
                if '/' not in rel_root:
                    new_role = False # Empty role directories are not new roles

        if new_role:
            return ([path + '/'], []) #new roles should just be synced all at once

        changed.sort(key=len)
        deleted = [p for p in prev_files.keys() if p.startswith(path) and p != path and p not in found_files] # also grab deleted files
        return (changed, deleted)


    def build(self, roles=None):
        '''Build a set of roles on this host, storing the snapshot of what will be done in $path/.build/'''
        unlock = self.lock()
        if roles is None:
            roles = self.roles
        
        self._owned = {}
        self._ready_packages = []
        self._ready_prepackage = {}
        shutil.rmtree(self.build_path, True)

        for role in roles:
            self.role(role).build()

        #For the sake of the build directory being a single, complete snapshot of what will be installed, serialize the package list to the metadata directory
        meta_path = os.path.join(self.build_path, '.bishop')
        try:
            os.makedirs(os.path.join(meta_path, '.prepackage'))
        except OSError:
            pass
        
        for name, script_path in self._ready_prepackage.items():
            os.symlink(os.path.abspath(script_path), os.path.join(meta_path, '.prepackage', name))

        open(os.path.join(meta_path, 'packages'), 'w').write('\n'.join(self._ready_packages))
        if unlock: self.unlock()

    def ready(self, build_path, role):
        '''Mark a built path as owned by the role that built it so we can detect and flag better errors on conflicts between roles'''
        self._owned[build_path] = role

    def ready_package(self, name):
        '''Add a package to the list of packages to be installed'''
        if not name in self._ready_packages:
            self._ready_packages.append(name)

    def ready_prepackage(self, path, name):
        '''Add a prepackage script to be run before a given package is installed'''
        self._ready_prepackage[name] = path

    def install(self, roles=None, save=True, path='/', local=False):
        '''Build the given roles, then install the build on the local host'''
        if self.install_locked and not local:
            raise Exception('install cancelled - local host is locked for testing')

        unlock = self.lock()
        if roles is None:
            roles = self.roles
        elif save:
            self.roles = list(set(self.roles).union(set(roles)))
        
        self._install_roles = roles
        # Roles should be added general-to-specific, but installed specific-to-general
        roles.reverse()

        if self.repo is not None:
            self.mirror(list(set(self.roles).union(set(roles))))
        self.build(roles)

        change_packages = self.get_new_packages()
        self.run_prepackages(change_packages)
        self.install_packages(change_packages)
        
        change_files = self.get_changes(path)
        self.run_preinstalls(change_files)
        [self.install_file(file, path) for file in change_files]
        self.run_postinstalls(change_files)
        
        self.last_install = time.time()
        self._install_roles = None
        self.save()
        if unlock: self.unlock()

    def uninstall(self, roles):
        '''Remove a set of roles from the roles this host checks'''
        unlock = self.lock()
        self.roles = list(set(self.roles).difference(set(roles)))
        self.save()
        if unlock: self.unlock()

    def get_new_packages(self):
        '''Return a list of only those packages that have not already been installed'''
        packager = package.Package()
        already_installed = packager.list()
        return [name for name in self._ready_packages if name not in already_installed]
        
    def run_prepackages(self, deps):
        '''Gather a unique list of the prepackage script to run based on what packages still need to be installed, and run them'''
        prepackages = {}
        for dep in deps:
            meta_path = os.path.join(self.build_path, '.bishop', '.prepackage', dep)
            if os.path.exists(meta_path):
                script_path = os.readlink(meta_path)
                if script_path in prepackages:
                    prepackages[script_path].append(dep)
                else:
                    prepackages[script_path] = [dep]

        for script_path, packages in self.script_order(prepackages):
            os.chmod(script_path, 0755)
            shell_args = [script_path]
            shell_args.extend(packages)
            logging.info('prepackage script %s' % script_path)
            util.shell(shell_args)

    def install_packages(self, deps):
        packager = package.Package()
        for dep in deps:
            logging.info('package install %s' % dep)
            packager.install(dep)

    def install_file(self, file_path, root):
        '''Atomically write the changed file to the given installation path'''
        install_path = os.path.join(root, file_path)
        if os.path.isfile(install_path):
            backup_path = os.path.join(self.path, '.backups', install_path[1:])
            if not os.path.isdir(os.path.dirname(backup_path)):
                try:
                    os.makedirs(os.path.dirname(backup_path))
                except:
                    pass
            shutil.copyfile(install_path, backup_path)
        build_path = os.path.join(self.build_path, file_path)
        build_path = os.path.abspath(build_path)

        #We assume that installations are allowed to mutate and move any files within the build directory
        orig_path = os.path.realpath(build_path)
        if os.path.isdir(orig_path) and not os.path.exists(install_path):
            util.chown_makedirs(install_path)
            return

        parent_dir = os.path.dirname(install_path)
        if not os.path.exists(parent_dir):
            util.chown_makedirs(parent_dir)

        # Pull the uid and gid from the parent directory
        st = os.stat(os.path.dirname(install_path))
        logging.info('install file %s (%s:%s)' % (install_path, st.st_uid, st.st_gid))

        if orig_path != build_path:
            # If the built file is a symlink, then copy the file so we can atomically move it to the install location without harming any non-build locations
            os.remove(build_path)
            shutil.copyfile(orig_path, build_path)
        shutil.move(build_path, install_path)
        os.chown(install_path, st.st_uid, st.st_gid)

    def run_preinstalls(self, changes):
        '''Scan the changed files looking for preinstall scripts and run them.  A preinstall script failure should terminate further installation.'''
        for script, files in self.script_order(self.pivot_meta(changes, 'preinstall')):
            os.chmod(script, 0755)
            shell_args = [script]
            shell_args.extend(['/%s' % name for name in files])
            logging.info('preinstall script %s' % script)
            util.shell(shell_args)

    def run_postinstalls(self, changes):
        '''Scan the changed files looking for postinstall scripts and run them.  A postinstall script failure should be logged but not halt running other postinstall scripts.'''
        postinstalls = self.pivot_meta(changes, 'postinstall')

        # Also include any postinstall scripts that have been modified since the last bishop run
        for role in [role for role in self._roles.itervalues() if role._built]:
            script_path = os.path.join(self.path, role.name, 'postinst')
            if os.path.exists(script_path) and script_path not in postinstalls and os.path.getmtime(script_path) > self.last_install:
                postinstalls[script_path] = []

        for script, files in self.script_order(postinstalls):
            os.chmod(script, 0755)
            shell_args = [script]
            shell_args.extend(['/%s' % name for name in files])
            logging.info('postinstall script %s' % script)
            try:
                util.shell(shell_args)
            except error.CalledProcessError, e:
                logging.error(str(e))

    def pivot_meta(self, changes, name):
        '''Scan all files in changes for metadata files tagged with {name}, then return a dictionary of {name} paths as keys and lists of files referencing the metadata file as the value'''
        meta = {}
        for file in changes:
            meta_path = '%s-%s' % (os.path.join(self.build_path, '.bishop', file), name)
            if os.path.exists(meta_path):
                script_path = os.readlink(meta_path)
                if script_path in meta:
                    meta[script_path].append(file)
                else:
                    meta[script_path] = [file]

        return meta

    def script_order(self, scripts):
        '''Take a dictionary of scripts: args and return an iterator for them in depth-first sorted order'''
        ordered = []
        for script, args in scripts.iteritems():
            ordered.append((script, args))

        if len(ordered) > 0:
            self.mark_role_order(self._install_roles)
            ordered.sort(key=lambda x: -self.role(os.path.basename(os.path.dirname(x[0])))._order)

        return ordered
    
    def get_changes(self, install_root='/'):
        '''Scan the build directories for files that are different than what's in the install root, using sha1 hashes to compare file contents'''
        changed = []
        self.merge_partials()

        def file_hash(path):
            h = sha()
            f = open(path, 'rb')
            chunk = f.read(524288)
            while chunk:
                h.update(chunk)
                chunk = f.read(524288)
            f.close()
            return h.hexdigest()

        if sys.version_info >= (2,6):
            files_to_walk = os.walk(self.build_path, followlinks=True)
        else:
            files_to_walk = os.walk(self.build_path)

        for root, dirs, files in files_to_walk:
            if '.bishop' in dirs:
                dirs.remove('.bishop')
            
            rel_root = root.replace(self.build_path, '')
            if rel_root.startswith('/'):
                rel_root = rel_root[1:]

            for file in files:
                file_path = os.path.join(root, file)
                install_path = os.path.join(install_root, rel_root, file)
                if not os.path.exists(install_path) or file_hash(file_path) != file_hash(install_path):
                    changed.append(os.path.join(rel_root, file))

            if len(files) == 0 and len(dirs) == 0 and rel_root != '':
                install_path = os.path.join(install_root, rel_root)
                # Create orphan directories and add them to the changed list so that the pre/post install scripts will trigger
                if not os.path.isdir(install_path):
                    changed.append(rel_root)

        return changed

    def extract_package(self, package_name, role_name):
        '''Extract the config files a given uninstalled package into a given role'''
        role_path = os.path.join(util.get_repopath(), role_name)
        package.Package().extract(package_name, os.path.join(role_path, 'files'))
        
        # Only keep the files in /etc (a simple way to isolate configuration from other stuff)
        for bad_path in [os.path.join(role_path, 'files', path) for path in os.listdir(os.path.join(role_path, 'files')) if path != 'etc']:
            shutil.rmtree(bad_path)
        
        # Write the package to the packages list (presume the dependencies will always come along for the ride)
        f = open(os.path.join(role_path, 'packages'), 'w')
        f.write(package_name + '\n')
        f.close()

    def lock(self):
        '''Lock this host for modification - prevent multiple installs or builds from running simultaneously, returning True if a new lock was acquired'''
        if self._locked:
            return False
        self._locked = open(os.path.join(self.path, '.lock'), 'w')
        fcntl.flock(self._locked, fcntl.LOCK_EX)
        return True

    def merge_partials(self):
        '''Scan the build directory for any defined partials and merge them into files'''
        meta_path = os.path.join(self.build_path, '.bishop')
        if sys.version_info >= (2,6):
            files_to_walk = os.walk(meta_path, followlinks=True)
        else:
            files_to_walk = os.walk(meta_path)

        for root, dirs, files in files_to_walk:
            rel_root = root.replace(meta_path, '')
            if rel_root.startswith('/'):
                rel_root = rel_root[1:]

            partials = {}
            for file in [f for f in files if '-bpartial-' in f]:
                file_path = os.path.join(root, file)
                name = os.path.join(rel_root, file.split('-bpartial-')[0])
                if name in partials:
                    partials[name].append(file_path)
                else:
                    partials[name] = [file_path]
            
            if len(partials) > 0:
                for name, paths in partials.items():
                    paths.sort()
                    partial_content = '\n'.join([open(path).read() for path in paths])
                    build_path = os.path.join(self.build_path, name)
                    if not os.path.exists(build_path):
                        try:
                            os.makedirs(os.path.dirname(build_path))
                        except OSError:
                            pass
                        open(build_path, 'w').write(partial_content)
                    else:
                        content = open(build_path).read()
                        os.remove(build_path)
                        open(build_path, 'w').write(content + '\n' + partial_content)

    def unlock(self):
        '''Unlock a given lock'''
        if not self._locked:
            return False

        fcntl.flock(self._locked, fcntl.LOCK_UN)
        self._locked.close()
        self._locked = False
        return True

class Remote(object):
    def __init__(self, host):
        if ':' not in host:
            host += ':4200'
        self._server = xmlrpclib.ServerProxy('http://%s' % host, allow_none=True)

    def __getattr__(self, attr):
        try:
            if hasattr(getattr(Local, attr), '__func__'):
                return getattr(self._server, attr)
            else:
                return getattr(self._server, attr)()
        except AttributeError:
            return getattr(self._server, attr)()


class Role(object):
    def __init__(self, repo, role):
        self._hosts = {}
        for host, info in repo.cluster[role].items():
            self._hosts[host] = Remote(info['ip'])

    def __getattr__(self, attr):
        def _getter(*args):
            ret = {}
            for host, server in self._hosts.items():
                ret[host] = getattr(server, attr)(*args)
            return ret
        
        try:
            if hasattr(getattr(Local, attr), '__func__'):
                return _getter
            else:
                return _getter()
        except AttributeError:
            return _getter()
        




class BuildError(Exception):
    pass
class RoleError(Exception):
    pass
class PackageError(Exception):
    pass
class ConflictError(BuildError):
    pass
class TemplateError(Exception):
    pass

class CalledProcessError(Exception):
    """This exception is raised when a process run by shell() returns
    a non-zero exit status.  The exit status will be stored in the
    returncode attribute and the stderr will be stored in stderr."""
    def __init__(self, returncode, cmd, stderr):
        self.returncode = returncode
        self.cmd = cmd
        self.stderr = stderr
    def __str__(self):
        return "Command '%s' returned non-zero exit status %d: %s" % (self.cmd, self.returncode, self.stderr)



from distutils.core import setup, Command
import os.path, shutil

class role_cmd(Command):
    description = 'install bishop into a directory as a bishop role'
    user_options = [
        ('role-dir=', 'd', 'directory to install the role into')
    ]

    def initialize_options(self):
        self.role_dir = './build.bishop.role'

    def finalize_options(self):
        parent_dir = os.path.dirname(self.role_dir)
        if parent_dir == '':
            parent_dir = '.'

        if not os.path.isdir(parent_dir):
            raise ValueError('Cannot save bishop role - "%s" is not a directory or does not exist' % parent_dir)

    def run(self):
        if not os.path.isdir(self.role_dir):
            os.makedirs(self.role_dir)
        
        src_root = os.path.join(self.role_dir, 'files/usr/local/src/bishop')
        if not os.path.isdir(src_root):
            os.makedirs(src_root)

        for path in ['bishop', 'scripts', 'data', 'setup.py']:
            dest = os.path.join(src_root, path)
            if os.path.isdir(dest):
                shutil.rmtree(dest)

            if os.path.isdir(path):
                shutil.copytree(path, os.path.join(src_root, path), ignore=lambda path, files: [file for file in files if file.endswith('.pyc')])
            else:
                shutil.copy(path, os.path.join(src_root, path))
        
        postinst = open(os.path.join(self.role_dir, 'postinst'), 'w')
        postinst.write('#!/bin/sh\n\ncd /usr/local/src/bishop\nchmod +x /usr/local/src/bishop/data/bishop-install\npython setup.py install\n')
        postinst.close()
        os.chmod(os.path.join(self.role_dir, 'postinst'), 0755)
        print 'Bishop copied as role to %s' % self.role_dir


setup(name='bishop', version='0.1.6', description='Simple Configuration Management',
      author='Chad Morris', author_email='chadwickmorris@gmail.com',
      license='MIT', keywords='configuration management',
      url='http://projects.chadwickmorris.com/bishop/', packages=['bishop', 'bishop.mako', 'bishop.mako.ext'],
      scripts=['scripts/bishop'], cmdclass={'install_role': role_cmd},
      data_files=[('/etc/cron.hourly', ['data/bishop-install'])],
      requires=['pexpect'],
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Topic :: System :: Installation/Setup'
      ]
)



import os.path, pipes, getpass, logging, pickle, tempfile, shutil, warnings
import util

class Repository(object):
    def __init__(self, path='/srv/bishop/'):
        self.path = path
        self.keyfile = None
        self.password = None
        if self.path[0] == '/' and not os.path.exists(self.path):
            os.makedirs(self.path)

        if not self.path.endswith('/'):
            self.path += '/'

    @property
    def all_roles(self):
        try:
            return [os.path.basename(path) for path in self.cmd('find {root}* -prune -type d').strip().splitlines()]
        except Exception, e:
            if 'No such file' in str(e):
                return []
            else:
                raise

    @property
    def hostname(self):
        '''return the host name part of repository location - or localhost for a local filesystem'''
        if self.path[0] == '/':
            return 'localhost'

        (server, path) = self.path.split(':', 1)
        if '@' in server:
            (user, server) = server.split('@', 1)

        return server

    @property
    def cluster(self):
        '''Grab the list of connected hosts and when they've installed, then return them split by roles'''
        warnings.simplefilter('ignore')
        path = os.tempnam() + '/'
        warnings.resetwarnings()
        os.mkdir(path)

        self.copy_down('.hosts.*', path)
        
        cluster = {}
        for role in self.all_roles:
            cluster[role] = {}
        
        for host_path in os.listdir(path):
            host_name = host_path[7:]
            saved = pickle.load(open(os.path.join(path, host_path), 'r'))
            for role in saved['roles']:
                if role in cluster:
                    cluster[role][host_name] = saved

            for role in saved['mixed_roles']:
                if role in cluster:
                    cluster[role]['(%s)' % host_name] = saved

        shutil.rmtree(path, True)
        return cluster

    @property
    def sync_disabled(self):
        '''Return False if the sync commands are enabled, or the message to show to users if they are disabled'''
        try:
            return self.cmd('cat {root}.bishop.nosync')
        except:
            return False

    def copy_down(self, from_path, to_path, include=None):
        '''Sync files from the repository to the given local filesystem path'''
        return self.copy(os.path.join(self.path, from_path), to_path, include)

    def copy_up(self, from_path, to_path, include=None):
        '''Sync files from the local filesystem to the repository'''
        return self.copy(from_path, os.path.join(self.path, to_path), include)

    def copy(self, from_path, to_path, include=None):
        '''Copy things around using rsync - presume that all paths are absolute or cwd-relative'''

        keyfile_auth = ''
        include_args = ''

        if self.keyfile is not None:
            keyfile_auth = '-e "ssh -i %s"' % pipes.quote(self.keyfile)

        if include is not None:
            include_args = ' '.join(['--include=%s' % pipes.quote(include_pat) for include_pat in include] + ['--exclude=*'])
        
        cmd = 'rsync -rlptDz --delete %s %s %s %s' % (keyfile_auth, include_args, from_path, to_path)
        
        if self.password is not None:
            return self.copy_passwd(cmd)
        else:
            return util.shell(cmd)

    def copy_passwd(self, cmd):
        '''Copy things around using rsync and pexpect to handle the tty-based password inputs'''
        import pexpect
        logging.debug(cmd)
        child = pexpect.spawn(cmd)
        child.expect('password: ')
        child.sendline(self.password)
        child.expect(pexpect.EOF)
        child.close()

    def delete_role(self, role):
        '''Delete a role from the repo'''
        return self.cmd('rm -rf {root}%s' % role)
    
    def disable_sync(self, msg=None):
        '''Disable the sync and sync-down commands, instead showing a message to the user'''
        f = tempfile.NamedTemporaryFile()
        content = 'The sync commands have been disabled on this repository.\n'
        if msg is not None:
            content += '\n' + msg + '\n'

        f.write(content)
        f.flush()
        os.chmod(f.name, 0644)
        self.copy_up(f.name, '.bishop.nosync')
        f.close()

    def enable_sync(self):
        '''Enable the sync and sync-down commands if they're disabled'''
        return self.cmd('rm -f {root}.bishop.nosync')

    def cmd(self, cmd):
        '''Execute an SSH command on the repo server'''
        if self.path[0] == '/':
            cmd = cmd.replace('{root}', self.path)
            return util.shell(cmd)

        if self.password is not None:
            return self.cmd_passwd(cmd)
        
        (server, path) = self.path.split(':', 1)
        cmd = cmd.replace('{root}', path)

        keyfile_auth = ''
        if self.keyfile is not None:
            keyfile_auth = '-i %s' % pipes.quote(self.keyfile)

        cmd = 'ssh %s %s %s' % (keyfile_auth, server, pipes.quote(cmd))
        return util.shell(cmd)

    def cmd_passwd(self, cmd):
        '''Execute an SSH command on the repo server, using pexpect to supply a password'''
        import pexpect
        
        (server, path) = self.path.split(':', 1)
        cmd = cmd.replace('{root}', path)

        cmd = 'ssh %s %s' % (server, pipes.quote(cmd))
        logging.debug(cmd)
        child = pexpect.spawn(cmd)
        child.expect('password: ')
        child.sendline(self.password)
        out = pexpect.read()
        child.close()
        return out

    def ask_password(self):
        '''Ask for the remote password without showing it in the tty'''
        self.password = getpass.getpass()

    def save(self, host):
        '''Maintain a log of hosts and roles so we can have a sense of the "cluster"'''
        to_save = ['roles', 'last_install', 'ip', 'mixed_roles']
        saved = {}
        for field in to_save:
            saved[field] = getattr(host, field)

        f = tempfile.NamedTemporaryFile()
        pickle.dump(saved, f)
        f.flush()
        os.chmod(f.name, 0644)
        self.copy_up(f.name, '.hosts.%s' % host.name)
        f.close()

    def __str__(self):
        return self.path
